import unittest

try:
	from StringIO import StringIO as BytesIO
except ImportError:
	from io import BytesIO

import datetime

import vesna.cdf
import vesna.cdf.xml

class TestCDFMetadata(unittest.TestCase):
	def test_empty(self):
		obj2 = vesna.cdf.xml._metadata_decode("")

		self.assertEqual(obj2, None)

	def test_basic(self):
		obj = ['foo']

		b = vesna.cdf.xml._metadata_encode(obj)
		obj2 = vesna.cdf.xml._metadata_decode(b)

		self.assertEqual(obj, obj2)

	def test_preserve(self):
		old_str = "hello, world!"

		obj = ['foo']

		b = vesna.cdf.xml._metadata_encode(obj, old_str)
		obj2 = vesna.cdf.xml._metadata_decode(b)

		self.assertEqual(obj, obj2)
		self.assertTrue(b.startswith(old_str))

		obj = ['bar']

		b = vesna.cdf.xml._metadata_encode(obj, b)
		obj2 = vesna.cdf.xml._metadata_decode(b)

		self.assertEqual(obj, obj2)
		self.assertTrue(b.startswith(old_str))

class TestCDFXMLExperiment(unittest.TestCase):
	def xest_create_save_load(self):
		e = vesna.cdf.CDFExperiment(title="test experiment", summary="test experiment",
				start_hz=1, stop_hz=10, step_hz=1)
		
		d = vesna.cdf.CDFDevice("http://localhost", 10000, 1)
		e.add_device(d)

		f = StringIO()
		e.save(f)

		f.seek(0)
		e2 = vesna.cdf.CDFExperiment.load(f)

		self.assertEqual(e.title, e2.title)
		self.assertEqual(e.summary, e2.summary)
		self.assertEqual(e.start_hz, e2.start_hz)
		self.assertEqual(e.stop_hz, e2.stop_hz)
		self.assertEqual(e.step_hz, e2.step_hz)
		self.assertEqual(e.tag, e2.tag)

		self.assertEqual(e.devices[0].base_url, e2.devices[0].base_url)
		self.assertEqual(e.devices[0].cluster_id, e2.devices[0].cluster_id)
		self.assertEqual(e.devices[0].addr, e2.devices[0].addr)

class TestCDFExperiment(unittest.TestCase):
	def test_create_iteration(self):
		i = vesna.cdf.CDFExperimentIteration()

	def create_experiment(self):

		e = vesna.cdf.CDFExperiment(
				title="Test experiment",
				summary="Experiment summary",
				release_date=datetime.datetime.now(),
				methodology="Collection methodology",
				related_experiments="Related experiments",
				notes="Notes")

		a = vesna.cdf.CDFAuthor(
				name="John",
				email="john@example.com",
				address="Nowhere",
				phone="0",
				institution="Institute of Imaginary Science")
		e.add_author(a)

		d = vesna.cdf.CDFDocument(
				description="A document",
				bibtex="BibTeX")
		e.add_document(d)

		e.set_frequency_range(
				start_hz=100e6,
				stop_hz=200e6,
				step_hz=1e6)

		e.set_duration(60)

		d = vesna.cdf.CDFDevice(
				base_url="http://example.com/communicator", 
				cluster_id=10000, 
				addr=1)

		e.add_device(d)

		d = vesna.cdf.CDFDevice(
				base_url="http://example.com/communicator", 
				cluster_id=10000, 
				addr=2)

		i = vesna.cdf.CDFInterferer(
				device=d)

		p = vesna.cdf.CDFInterfererProgram(
				center_hz=150e6,
				power_dbm=0,

				device_id=0,
				config_id=0,

				start_time=10,
				end_time=20)

		i.add_program(p)

		e.add_interferer(i)

		return e

	def test_create(self):
		self.create_experiment()

	def test_xml_create(self):
		e = self.create_experiment()
		vesna.cdf.xml.CDFXMLExperiment(e)

	def test_xml_save_load(self):
		io = BytesIO()

		e = self.create_experiment()
		e = vesna.cdf.xml.CDFXMLExperiment(e)
		e.save(io)

		#print io.getvalue()

		io.seek(0)

		e2 = vesna.cdf.xml.CDFXMLExperiment.load(io)

		self.assertEqual(e.get_experiment().title, e2.get_experiment().title)



import unittest

from vesna.alh import CRCError
from vesna.alh import signalgenerator

#import logging
#logging.basicConfig(level=logging.DEBUG)

class TestSignalGenerator(unittest.TestCase):

	def test_get_config_list(self):

		class TestALH:
			def get(self, resource):
				s = "dev #0, Test, 1 configs:\n" \
					"  cfg #0: Test:\n" \
					"     base: 10 Hz, spacing: 1 Hz, bw: 1 Hz, channels: 10, min power: -10 dBm, max power: 0 dBm, time: 1 ms"
				return s.encode('ascii')

		alh = TestALH()
		s = signalgenerator.SignalGenerator(alh)

		cl = s.get_config_list()

		self.assertEqual(len(cl.devices), 1)
		self.assertEqual(len(cl.configs), 1)

	def test_get_config_list_corrupt_1(self):

		class TestALH:
			def get(self, resource):
				return b""

		alh = TestALH()
		s = signalgenerator.SignalGenerator(alh)

		cl = s.get_config_list()

		self.assertEqual(cl.configs, [])
		self.assertEqual(cl.devices, [])

	def test_get_config_list_corrupt_2(self):

		class TestALH:
			def get(self, resource):
				return b"dev #0, Test, 2 configs:"

		alh = TestALH()
		s = signalgenerator.SignalGenerator(alh)

		self.assertRaises(CRCError, s.get_config_list)

	def test_get_config_list_corrupt_3(self):

		class TestALH:
			def get(self, resource):
				s = "dev #0, Test, 1 configs:\n"\
					"  cfg #0: Test:"
				return s.encode('ascii')

		alh = TestALH()
		s = signalgenerator.SignalGenerator(alh)

		self.assertRaises(CRCError, s.get_config_list)


class TestGeneratorConfigList(unittest.TestCase):
	def test_get_config_name(self):
		cl = signalgenerator.ConfigList()

		d = signalgenerator.Device(0, "test")
		cl._add_device(d)

		def add_dc(id, name, base):
			dc = signalgenerator.DeviceConfig(id, name, d)
			dc.base = base 
			dc.spacing = 1
			dc.num = 1000
			dc.time = 1
			dc.min_power = -100
			dc.max_power = 0
			cl._add_config(dc)

		add_dc(0, "foo 1", 1000)
		add_dc(1, "foo 2", 2000)
		add_dc(2, "bar 1", 1000)
		add_dc(3, "bar 2", 2000)

		sc = cl.get_tx_config(1500, 0)
		self.assertEqual(0, sc.config.id)

		sc = cl.get_tx_config(2500, 0)
		self.assertEqual(1, sc.config.id)

		sc = cl.get_tx_config(1500, 0, name="bar")
		self.assertEqual(2, sc.config.id)

		sc = cl.get_tx_config(2500, 0, name="bar")
		self.assertEqual(3, sc.config.id)

from vesna.alh.spectrumsensor import SpectrumSensor, SpectrumSensorResult, SpectrumSensorProgram
from vesna.spectrumsensor import Device, DeviceConfig, SweepConfig, Sweep

class TestSpectrumSensor(unittest.TestCase):

	def test_get_config_list(self):

		class TestALH:
			def get(self, resource):
				s =	"dev #0, Test, 1 configs:\n" \
					"  cfg #0: Test:\n" \
					"     base: 10 Hz, spacing: 1 Hz, bw: 1 Hz, channels: 10, time: 1 ms"
				return s.encode('ascii')

		alh = TestALH()
		s = SpectrumSensor(alh)

		cl = s.get_config_list()

		self.assertEqual(len(cl.devices), 1)
		self.assertEqual(len(cl.configs), 1)

	def test_get_config_list_corrupt_1(self):

		class TestALH:
			def get(self, resource):
				return "".encode('ascii')

		alh = TestALH()
		s = SpectrumSensor(alh)

		cl = s.get_config_list()

		self.assertEqual(cl.configs, [])
		self.assertEqual(cl.devices, [])

	def test_get_config_list_corrupt_2(self):

		class TestALH:
			def get(self, resource):
				s = "dev #0, Test, 2 configs:"
				return s.encode('ascii')

		alh = TestALH()
		s = SpectrumSensor(alh)

		self.assertRaises(CRCError, s.get_config_list)

	def test_get_config_list_corrupt_3(self):

		class TestALH:
			def get(self, resource):
				s = "dev #0, Test, 1 configs:\n"\
					"  cfg #0: Test:"
				return s.encode('ascii')

		alh = TestALH()
		s = SpectrumSensor(alh)

		self.assertRaises(CRCError, s.get_config_list)

	def _get_sc(self):
		d = Device(0, "test")

		dc = DeviceConfig(0, "foo", d)
		dc.base = 1000
		dc.spacing = 1
		dc.num = 1000
		dc.time = 1

		sc = SweepConfig(dc, 0, 3, 1)

		return sc

	def test_sweep_1(self):
		class MockALH(ALHProtocol):
			def _post(self, resource, data, *args):
				return b"\x00\x00\x01\x00\x02\x00D\xa4H;"

		alh = MockALH()
		ss = SpectrumSensor(alh)

		sc = self._get_sc()
		r = ss.sweep(sc)

		self.assertEqual(r.data, [0., .01, .02])

	def test_sweep_2(self):
		class MockALH(ALHProtocol):
			def _post(self, resource, data, *args):
				# negative CRC
				return b"\x00\x00\x01\x00\x08\x00\xceL\xa7\xc1"

		alh = MockALH()
		ss = SpectrumSensor(alh)

		sc = self._get_sc()
		r = ss.sweep(sc)

		self.assertEqual(r.data, [0., .01, .08])

	def test_retrieve(self):
		class MockALH(ALHProtocol):
			def _get(self, resource, *args):
				if "Info" in resource:
					return b"status=COMPLETE,size=14"
				else:
					return b"\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x91m\x00i"

		alh = MockALH()
		ss = SpectrumSensor(alh)

		sc = self._get_sc()
		p = SpectrumSensorProgram(sc, 0, 10, 1)

		r = ss.retrieve(p)

		self.assertEqual(len(r.sweeps), 1)
		self.assertEqual(r.sweeps[0].data, [0., .01, .02])
		self.assertEqual(r.sweeps[0].timestamp, 0)

import tempfile

class TestSpectrumSensorResult(unittest.TestCase):
	def setUp(self):
		d = Device(0, "test")

		dc = DeviceConfig(0, "foo", d)
		dc.base = 1000
		dc.spacing = 1
		dc.num = 1000

		sc = SweepConfig(dc, 0, 3, 1)

		p = SpectrumSensorProgram(sc, 0, 10, 1)

		self.r = r = SpectrumSensorResult(p)

		s = Sweep()
		s.timestamp = 0.0
		s.data = [ 0.0, 1.0, 2.0 ]
		r.sweeps.append(s)

		s = Sweep()
		s.timestamp = 1.0
		s.data = [ 3.0, 4.0 ]
		r.sweeps.append(s)

	def test_get_data(self):
		self.assertEqual( [
			[ 0.0, 1.0, 2.0 ],
			[ 3.0, 4.0, 4.0 ]
		], self.r.get_data())

	def test_get_hz_list(self):
		self.assertEqual( [ 1000.0, 1001.0, 1002.0 ], self.r.get_hz_list() )

	def test_get_s_list(self):
		self.assertEqual( [ 0.0, 1.0 ], self.r.get_s_list() )

	def test_write(self):
		f = tempfile.NamedTemporaryFile()

		self.r.write(f.name)

		f.seek(0)
		a = f.read()

		self.assertEqual(a, b"""# t [s]	f [Hz]	P [dBm]
0.000000	1000.000000	0.000000
0.333333	1001.000000	1.000000
0.666667	1002.000000	2.000000

1.000000	1000.000000	3.000000
1.333333	1001.000000	4.000000

""")

from vesna.alh import ALHProtocol

class TestALHProtocol(unittest.TestCase):
	def test_is_printable_1(self):
		self.assertEqual(ALHProtocol._is_printable("foo".encode('ascii')), True)

	def test_is_printable_2(self):
		self.assertEqual(ALHProtocol._is_printable("\x00".encode('ascii')), False)

	def test_is_printable_3(self):
		self.assertEqual(ALHProtocol._is_printable(u"\x8f".encode('latin2')), False)

try:
	from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
except ImportError:
	from http.server import HTTPServer, BaseHTTPRequestHandler

import threading
from vesna.alh import ALHWeb

class TestALHWeb(unittest.TestCase):
	def setUp(self):

		self.last_headers = last_headers = [None]
		self.srv_response = srv_response = [None]

		class MockHTTPRequestHandler(BaseHTTPRequestHandler):
			def do_GET(self):
				last_headers[0] = self.headers

				self.send_response(200)
				self.end_headers()

				self.wfile.write(srv_response[0])

			def log_message(self, format, *args):
				pass

		server_address = ('localhost', 12345)
		self.httpd = HTTPServer(server_address, MockHTTPRequestHandler)
		self.t = threading.Thread(target=self.httpd.serve_forever)
		self.t.start()

	def tearDown(self):
		self.httpd.shutdown()
		self.t.join()
		self.httpd.server_close()

	def test_get_ascii(self):
		self.srv_response[0] = 'bar'.encode('ascii')

		alh = ALHWeb("http://localhost:12345", "id")
		r = alh.get("foo")

		self.assertEqual(r, self.srv_response[0])
		self.assertIn('alh', self.last_headers[0]['User-Agent'])

	def test_get_bin(self):
		self.srv_response[0] = '\x00\x01\x02\x04'.encode('ascii')

		alh = ALHWeb("http://localhost:12345", "id")
		r = alh.get("foo")

		self.assertEqual(r, self.srv_response[0])



import logging
from vesna import alh
import serial
import sys

def add_communication_options(parser):
	parser.add_option("-U", "--url", dest="url", metavar="URL",
			help="Use URL for communication with coordinator")
	parser.add_option("-u", "--cluster", dest="cluster_id", metavar="ID", type="int",
			help="Cluster ID to pass to the web API")

	parser.add_option("-D", "--device", dest="device", metavar="PATH",
			help="Use serial terminal for communication with coordinator")
	parser.add_option("-v", "--verbosity", dest="verbosity", metavar="LEVEL",
			help="Set verbosity level (debug, info, warning, error)")

def get_coordinator(options):
	if options.url and not options.device:
		coordinator = alh.ALHWeb(options.url, options.cluster_id)
	elif options.device and not options.url:
		f = serial.Serial(options.device, 115200, timeout=180)
		coordinator = alh.ALHTerminal(f)
	else:
		raise Exception("Please give either -U or -D option")

	if options.verbosity:
		level = getattr(logging, options.verbosity.upper(), None)
		if not isinstance(level, int):
			raise Exception("Invalid verbosity level: %s" % (level,))
	else:
		level = logging.INFO

	logging.basicConfig(level=level)

	return coordinator



import datetime
import dateutil.parser
from lxml import etree
import json
import os.path
import time
import uuid

from vesna import cdf

_METADATA_HEADER = "Additional VESNA metadata follows:\n\n"

def _metadata_encode(obj, string=''):
	i = string.find(_METADATA_HEADER)
	if i != -1:
		string = string[:i]

	return string + _METADATA_HEADER + json.dumps(obj, indent=4)

def _metadata_decode(string):
	i = string.find(_METADATA_HEADER)
	if i != -1:
		return json.loads(string[i+len(_METADATA_HEADER):])

def text_or_none(xml_tree, xpath):
	t = xml_tree.find(xpath)
	if t is not None:
		return t.text
	else:
		return None

class CDFXMLExperiment:
	def __init__(self, experiment):
		self.exp = experiment

	def get_experiment(self):
		return self.exp

	@classmethod
	def load(cls, f):
		root = etree.parse(f)

		exp = cls._from_xml(root)

		return cls(exp)

	@classmethod
	def _from_xml(cls, root):
		title = text_or_none(root, "experimentAbstract/title")

		tag = text_or_none(root, "experimentAbstract/uniqueCREWTag")

		summary = text_or_none(root, "experimentAbstract/experimentSummary")

		release_date = text_or_none(root, "experimentAbstract/releaseDate")
		release_date = dateutil.parser.parse(release_date)

		methodology = []
		for method in root.findall("experimentAbstract/collectionMethodology"):
			methodology.append(method.text)

		related_experiments = text_or_none(root, "experimentAbstract/relatedExperiments")

		extra = None

		notes = []
		for note in root.findall("experimentAbstract/notes"):

			o = _metadata_decode(note.text)
			if o:
				extra = o
			else:
				notes.append(note.text)

		experiment = cdf.CDFExperiment(
				title=title,
				summary=summary,
				release_date=release_date,
				methodology=methodology,
				related_experiments=related_experiments,
				notes=notes)


		for author in root.findall("experimentAbstract/author"):
			experiment.add_author(cls._author_from_xml(author))

		for document in root.findall("experimentAbstract/furtherDocumentation"):
			experiment.add_document(cls._document_from_xml(document))


		start_hz = float(text_or_none(root, "metaInformation/radioFrequency/startFrequency"))
		stop_hz = float(text_or_none(root, "metaInformation/radioFrequency/stopFrequency"))
		step_hz = extra['step_hz']

		experiment.set_frequency_range(start_hz, stop_hz, step_hz)


		duration = extra['duration']
		experiment.set_duration(duration)


		devices = {}
		for d in root.findall("metaInformation/device"):
			device = cls._device_from_xml(d)
			devices[device.key()] = device

		interference = root.find("metaInformation/radioFrequency/interferenceSources")
		cls._interferers_from_xml(interference, experiment, devices)

		for device in devices.values():
			experiment.add_device(device)

		return experiment

	@classmethod
	def _author_from_xml(cls, root):
		
		name = text_or_none(root, "name")
		email = text_or_none(root, "email")

		address = []
		for t in root.findall("address"):
			address.append(t.text)

		phone = []
		for t in root.findall("phone"):
			phone.append(t.text)

		institution = []
		for t in root.findall("institution"):
			institution.append(t.text)

		return cdf.CDFAuthor(
				name=name,
				email=email,
				address=address,
				phone=phone,
				institution=institution)

	@classmethod
	def _document_from_xml(cls, root):
		
		description = []
		for t in root.findall("description"):
			description.append(t.text)

		bibtex = []
		for t in root.findall("bibtex"):
			bibtex.append(t.text)

		return cdf.CDFDocument(
				description=description,
				bibtex=bibtex)

	@classmethod
	def _device_from_xml(cls, root):
		extra = _metadata_decode(text_or_none(root, "description"))
		return cdf.CDFDevice(**extra)

	@classmethod
	def _interferers_from_xml(cls, root, experiment, devices):
		if not root.text:
			return

		extra_interferers = _metadata_decode(root.text)

		for extra_interferer in extra_interferers['interferers']:

			device_key = tuple(extra_interferer['device'])
			device = devices.pop(device_key)

			interferer = cdf.CDFInterferer(device=device)

			experiment.add_interferer(interferer)

			for program in extra_interferer['programs']:
				interferer.add_program(cdf.CDFInterfererProgram(**program))

	def _format_date(self, date):
		return date.isoformat()

	def _to_xml(self):
		root = etree.Element("experimentDescription")


		abstract = etree.SubElement(root, "experimentAbstract")

		title = etree.SubElement(abstract, "title")
		title.text = self.exp.title

		tag = etree.SubElement(abstract, "uniqueCREWTag")
		tag.text = self.exp.tag

		for author in self.exp.authors:
			abstract.append(self._author_to_xml(author))

		date = etree.SubElement(abstract, "releaseDate")
		date.text = self._format_date(self.exp.release_date)

		summary = etree.SubElement(abstract, "experimentSummary")
		summary.text = self.exp.summary

		for t in self.exp.methodology:
			method = etree.SubElement(abstract, "collectionMethodology")
			method.text = t

		for document in self.exp.documentation:
			abstract.append(self._document_to_xml(document))

		related = etree.SubElement(abstract, "relatedExperiments")
		related.text = self.exp.related_experiments

		ext_note = _metadata_encode({
			"duration": self.exp.duration,
			"step_hz": self.exp.step_hz})

		for t in self.exp.notes + [ext_note]:
			note = etree.SubElement(abstract, "notes")
			note.text = t


		meta = etree.SubElement(root, "metaInformation")

		for device in self.exp.iter_all_devices():
			meta.append(self._device_to_xml(device))

		location = etree.SubElement(meta, "location")

		# FIXME
		layout = etree.SubElement(location, "layout")
		mobility = etree.SubElement(location, "mobility")

		# FIXME
		date = etree.SubElement(meta, "date")
		date.text = self._format_date(datetime.datetime.now())

		rf = etree.SubElement(meta, "radioFrequency")

		start = etree.SubElement(rf, "startFrequency")
		start.text = str(self.exp.start_hz)

		stop = etree.SubElement(rf, "stopFrequency")
		stop.text = str(self.exp.stop_hz)

		rf.append(self._interferers_to_xml())

		trace = etree.XML("""<traceDescription><format>Tab-separated-values file with timestamp, frequency, power triplets.</format><fileFormat><header>Comment line, starting with #</header><collectedMetrics><name>time</name><unitOfMeasurements>s</unitOfMeasurements></collectedMetrics><collectedMetrics><name>frequency</name><unitOfMeasurements>Hz</unitOfMeasurements></collectedMetrics><collectedMetrics><name>power</name><unitOfMeasurements>dBm</unitOfMeasurements></collectedMetrics></fileFormat></traceDescription>""")

		meta.append(trace)


		for iteration in self.exp.iterations:
			if iteration.start_time:
				root.append(self._iteration_to_xml(iteration))

		tree = etree.ElementTree(root)
		return tree

	def _author_to_xml(self, author):
		root = etree.Element("author")

		name = etree.SubElement(root, "name")
		name.text = author.name

		email = etree.SubElement(root, "email")
		email.text = author.email

		for t in author.address:
			address = etree.SubElement(root, "address")
			address.text = t

		for t in author.phone:
			phone = etree.SubElement(root, "phone")
			phone.text = t

		for t in author.institution:
			institution = etree.SubElement(root, "institution")
			institution.text = t

		return root

	def _document_to_xml(self, document):
		root = etree.Element("furtherDocumentation")

		for t in document.description:
			description = etree.SubElement(root, "description")
			description.text = t

		for t in document.bibtex:
			bibtex = etree.SubElement(root, "bibtex")
			bibtex.text = t

		return root

	def _device_to_xml(self, device):
		root = etree.Element("device")

		name = etree.SubElement(root, "name")
		name.text = "VESNA node %d" % device.addr

		description = etree.SubElement(root, "description")
		description.text = _metadata_encode({
			"base_url": device.base_url,
			"cluster_id": device.cluster_id,
			"addr": device.addr})

		return root

	def _iteration_to_xml(self, iteration):
		root = etree.Element("experimentIteration")

		time = etree.SubElement(root, "time")

		start = etree.SubElement(time, "starttime")
		start.text = self._format_date(iteration.start_time)

		end = etree.SubElement(time, "endtime")
		end.text = self._format_date(iteration.end_time)

		for tracefile in iteration.tracefiles:
			path = etree.SubElement(root, "traceFile")
			path.text = tracefile

		return root

	def _interferers_to_xml(self):

		i_list = []
		for interferer in self.exp.interferers:
			p_list = []
			for program in interferer.programs:
				p_struct = {
					'center_hz': program.center_hz,
					'power_dbm': program.power_dbm,
					'start_time': program.start_time,
					'end_time': program.end_time }
				p_list.append(p_struct)

			i_struct = {
				'device': interferer.device.key(),
				'programs': p_list }

			i_list.append(i_struct)

		interference = etree.Element("interferenceSources")
		interference.text = _metadata_encode({"interferers": i_list})

		return interference

	def save(self, f):
		tree = self._to_xml()
		tree.write(f, pretty_print=True, encoding='utf8')

	def save_all(self, path=None):
		if path is None:
			path = self.exp.tag

		cdf_path = path + ".cdf"
		dat_path = path + ".dat"

		try:
			os.mkdir(dat_path)
		except OSError:
			pass

		for iteration in self.exp.iterations:
			for i, sensor in enumerate(iteration.sensors):

				n = "data_%s_node_%d_%d.dat" % (
						iteration.start_time.strftime("%Y%m%d"),
						sensor.sensor.alh.addr,
						i)
				p = os.path.join(dat_path, n)

				sensor.result.write(p)

				iteration.tracefiles.append(p)

		self.save(open(cdf_path, "w"))



import math
import time

from vesna.alh import ALHProxy
from vesna.alh.common import add_communication_options, get_coordinator
from vesna.alh.spectrumsensor import SpectrumSensor, SpectrumSensorProgram, SweepConfig
from vesna.rftest import DeviceUnderTest

class RemoteDeviceUnderTest(DeviceUnderTest):
	def add_options(self, parser):
		add_communication_options(parser)

		parser.add_option("-n", "--node", dest="node", metavar="ADDR", type="int",
				help="Connect to node with ZigBit address ADDR")

	def setup(self, options):

		if not options.verbosity:
			options.verbosity = "warning"

		coor = get_coordinator(options)
		coor.post("prog/firstcall", "1")

		self.node = ALHProxy(coor, options.node)
		self.node.post("prog/firstcall", "1")

		self.spectrumsensor = SpectrumSensor(self.node)

		self.config_list = self.spectrumsensor.get_config_list()
		if not self.config_list.configs:
			raise Exception("Device returned no configurations. "
					"It is still scanning or not responding.")

		self.config = self.config_list.get_config(self.device_id, self.config_id)

	def get_fw_version(self):
		return self.node.get("hello").strip()

	def get_status(self):
		resp = self.node.get("sensing/deviceStatus").strip()
		return [v.strip() for v in resp.split("\n") ]

	def measure_ch_impl(self, ch, n):
		sweep_config = SweepConfig(self.config, ch, ch+1, 1)

		duration = int(math.ceil(self.config.time * n * 1e-3 + 1.0))

		now = time.time()
		sensor_program = SpectrumSensorProgram(sweep_config, now + 1, duration, 2)

		self.spectrumsensor.program(sensor_program)
		while not self.spectrumsensor.is_complete(sensor_program):
			time.sleep(1)

		result = self.spectrumsensor.retrieve(sensor_program)

		measurements = [ sweep.data[0] for sweep in result.sweeps ]
		measurements = measurements[:n]

		assert len(measurements) == n
		return measurements





import binascii
try:
	from itertools import zip_longest
except ImportError:
	from itertools import izip_longest as zip_longest
import logging
import re
import struct
import time

from vesna.spectrumsensor import Device, DeviceConfig, ConfigList, SweepConfig, Sweep
from vesna.alh import CRCError, ALHException

log = logging.getLogger(__name__)

class ALHProgrammingTimeError(ALHException): pass

class SpectrumSensorProgram:
	"""Describes a single spectrum sensing task.

	:param sweep_config: frequency sweep configuration to use, a :py:class:`SweepConfig` object
	:param time_start: time to start the task (UNIX timestamp)
	:param time_duration: duration of the task in seconds
	:param slot_id: numerical slot id used for storing measurements
	"""

	def __init__(self, sweep_config, time_start, time_duration, slot_id):
		self.sweep_config = sweep_config
		self.time_start = time_start
		self.time_duration = time_duration
		self.slot_id = slot_id

class SpectrumSensorResult:
	"""Result of a spectrum sensing task."""

	def __init__(self, program):
		"""Create a new result object.

		program -- SpectrumSensorProgram object that produced these results.
		"""
		self.program = program
		self.sweeps = []

	def get_hz_list(self):
		"""Return a list of frequencies in hertz covered by this result.
		"""
		return self.program.sweep_config.get_hz_list()

	def get_s_list(self):
		"""Return a list of timestamps in seconds covered by this result.
		"""
		return [ sweep.timestamp for sweep in self.sweeps ]

	def get_data(self):
		"""Return power measurements in dbm in form a two-dimensional array.
		"""
		data = []

		row_len = len(self.program.sweep_config.get_ch_list())

		for sweep in self.sweeps:
			if len(sweep.data) == row_len:
				row = sweep.data
			else:
				# only last row can be shorter
				assert len(data) == len(self.sweeps) - 1

				row = sweep.data + [sweep.data[-1]] * (row_len - len(sweep.data))

			data.append(row)

		return data

	def write(self, path):
		"""Write measurements into a tab-separated-values file.

		:param path: path to the file to write
		"""

		outf = open(path, "w")

		outf.write("# t [s]\tf [Hz]\tP [dBm]\n")

		sweep_config = self.program.sweep_config
		num_channels = sweep_config.num_channels
		sweep_time = 0.0

		next_sweep_i = iter(self.sweeps)
		next(next_sweep_i)
		i = zip_longest(self.sweeps, next_sweep_i)

		for sweepnr, (sweep, next_sweep) in enumerate(i):
			assert isinstance(sweep, Sweep)

			if next_sweep is not None:
				sweep_time = next_sweep.timestamp - sweep.timestamp

			for dbmn, dbm in enumerate(sweep.data):

				time = sweep.timestamp + sweep_time/num_channels * dbmn

				channel = sweep_config.start_ch + sweep_config.step_ch * dbmn
				assert channel < sweep_config.stop_ch

				freq = sweep_config.config.ch_to_hz(channel)

				outf.write("%f\t%f\t%f\n" % (time, freq, dbm))

			outf.write("\n")

		outf.close()

class SpectrumSensor:
	"""ALH node acting as a spectrum sensor.

	:param alh: ALH implementation used to communicate with the node
	"""
	MAX_TIME_ERROR = 2.0
	MAX_SINGLE_SWEEP_TIME = 800e-3

	def __init__(self, alh):
		self.alh = alh

	def _split_sweep_config(self, sweep_config):

		ch_per_sweep = int(self.MAX_SINGLE_SWEEP_TIME / (sweep_config.config.time * 1e-3))

		sweep_config_list = []

		start_ch = sweep_config.start_ch
		step_ch = sweep_config.step_ch
		while start_ch < sweep_config.stop_ch:
			stop_ch = min(sweep_config.stop_ch, start_ch + ch_per_sweep * step_ch)

			sweep_config_list.append(SweepConfig(sweep_config.config, start_ch, stop_ch, step_ch))
			start_ch = stop_ch

		return sweep_config_list

	@staticmethod
	def _crc32(data):
		v= binascii.crc32(data) & 0xffffffff
		return v

	def _sweep(self, sweep_config):
		response = self.alh.post("sensing/quickSweepBin",
				"dev %d conf %d ch %d:%d:%d" % (
				sweep_config.config.device.id,
				sweep_config.config.id,
				sweep_config.start_ch,
				sweep_config.step_ch,
				sweep_config.stop_ch))

		data = response[:-4]
		crc = response[-4:]

		their_crc = struct.unpack("<I", crc[-4:])[0]
		our_crc = self._crc32(data)
		if their_crc != our_crc:
			# Firmware versions 2.29 only calculate CRC on the
			# first half of the response due to a bug
			our_crc = self._crc32(data[:len(data)//2])
			if their_crc != our_crc:
				raise CRCError
			else:
				log.warning("working around broken CRC calculation! "
						"please upgrade node firmware")

		assert sweep_config.num_channels * 2 == len(data)

		result = []
		for n in range(0, len(data), 2):
			datum = data[n:n+2]

			dbm = struct.unpack("<h", datum)[0]*1e-2
			result.append(dbm)

		return result

	def sweep(self, sweep_config):
		"""Perform a single frequency sweep and return results
		immediately

		:param sweep_config: frequency sweep configuration to use, a :py:class:`SweepConfig` object
		"""

		sweep = Sweep()
		sweep.timestamp = 0

		for sweep_config in self._split_sweep_config(sweep_config):
			data = self._sweep(sweep_config)
			sweep.data += data

		return sweep

	def program(self, program):
		"""Send the given spectrum sensing program to the node.

		:param program: a :py:class:`SpectrumSensorProgram` object
		"""

		self.alh.post("sensing/freeUpDataSlot", "1", "id=%d" % (program.slot_id))

		time_before = time.time()

		relative_time = int(program.time_start - time_before)
		if relative_time < 0:
			raise Exception("Start time can't be in the past")

		self.alh.post("sensing/program",
			"in %d sec for %d sec with dev %d conf %d ch %d:%d:%d to slot %d" % (
				relative_time,
				program.time_duration,
				program.sweep_config.config.device.id,
				program.sweep_config.config.id,
				program.sweep_config.start_ch,
				program.sweep_config.step_ch,
				program.sweep_config.stop_ch,
				program.slot_id))

		time_after = time.time()

		time_error = time_after - time_before
		if time_error > self.MAX_TIME_ERROR:
			raise ALHProgrammingTimeError("Programming time error %.1f s > %.1fs" % 
					(time_error, self.MAX_TIME_ERROR))

	def is_complete(self, program):
		"""Return true if given program has been successfuly completed.

		:param program: a :py:class:`SpectrumSensorProgram` object
		"""
		if time.time() < program.time_start + program.time_duration:
			return False
		else:
			resp = self.alh.get("sensing/slotInformation", "id=%d" % (program.slot_id,))
			return "status=COMPLETE" in resp.decode("UTF-8")

	@staticmethod
	def _decode(program, data):
		num_channels = program.sweep_config.num_channels
		line_bytes = num_channels * 2 + 4

		result = SpectrumSensorResult(program)

		sweep = Sweep()
		for n in range(0, len(data), 2):
			datum = data[n:n+2]
			if len(datum) != 2:
				continue

			if n % line_bytes == 0:
				# got a time-stamp
				t = data[n:n+4]
				tt = struct.unpack("<i", t)[0]
				assert not sweep.data
				sweep.timestamp = tt * 1e-3
				continue

			if n % line_bytes == 2:
				# second part of a time-stamp, just ignore
				assert not sweep.data
				continue

			dbm = struct.unpack("<h", datum)[0]*1e-2
			sweep.data.append(dbm)

			if len(sweep.data) >= num_channels:
				result.sweeps.append(sweep)
				sweep = Sweep()

		if(sweep.data):
			result.sweeps.append(sweep)

		return result

	def retrieve(self, program):
		"""Retrieve results from the given spectrum sensing program.

		:param program: a :py:class:`SpectrumSensorProgram` object
		:return: a :py:class:`SpectrumSensorResult` object
		"""
		resp = self.alh.get("sensing/slotInformation", "id=%d" % (program.slot_id,))
		resp_ascii = resp.decode('ascii')

		assert "status=COMPLETE" in resp_ascii

		g = re.search("size=([0-9]+)", resp_ascii)
		total_size = int(g.group(1))

		#print "total size:", total_size

		p = 0
		max_read_size = 512
		data = b""

		while p < total_size:
			chunk_size = min(max_read_size, total_size - p)

			#if p < total_size - chunk_size*2:
			#	p += max_read_size
			#	continue

			#print "start", p
			#print "size", chunk_size

			chunk_data_crc = self.alh.get("sensing/slotDataBinary", "id=%d&start=%d&size=%d" % (
				program.slot_id, p, chunk_size))

			chunk_data = chunk_data_crc[:-4]

			#print "len", len(chunk_data)
			
			their_crc = struct.unpack("I", chunk_data_crc[-4:])[0]
			our_crc = self._crc32(chunk_data)

			if(their_crc != our_crc):
				raise CRCError

			data += chunk_data

			p += max_read_size

		return self._decode(program, data)

	def get_config_list(self):
		"""Query and return the list of supported device configurations.

		:return: a :py:class:`ConfigList` object
		"""
		config_list = ConfigList()

		device = None
		config = None

		description = self.alh.get("sensing/deviceConfigList")
		description = description.decode("ascii")
		configs_left = 0
		state = 0
		for line in description.split("\n"):
			g = re.match("dev #([0-9]+), (.+), ([0-9]+) configs:", line)
			if state == 0 and g:
				device = Device(int(g.group(1)), g.group(2))
				config_list._add_device(device)
				configs_left = int(g.group(3))
				state = 1
				continue

			g = re.match("  cfg #([0-9]+): (.+):", line)
			if state == 1 and g:
				config = DeviceConfig(int(g.group(1)), g.group(2), device)
				state = 2

				continue

			g = re.match("     base: ([0-9]+) Hz, spacing: ([0-9]+) Hz, bw: ([0-9]+) Hz, channels: ([0-9]+), time: ([0-9]+) ms", line)
			if state == 2 and g:
				config.base = int(g.group(1))
				config.spacing = int(g.group(2))
				config.bw = int(g.group(3))
				config.num = int(g.group(4))
				config.time = int(g.group(5))

				config_list._add_config(config)

				configs_left -= 1
				if configs_left < 0:
					raise CRCError
				elif configs_left == 0:
					state = 0
				else:
					state = 1

				continue

		if configs_left != 0:
			raise CRCError

		return config_list



import binascii
import logging
import os
import re
import string
import sys
import time
import ssl

try:
	# Python 2.x
	from urllib import FancyURLopener, urlencode
except ImportError:
	# Python 3.x
	from urllib.request import FancyURLopener
	from urllib.parse import urlencode

log = logging.getLogger(__name__)

class ALHException(Exception):
	"""Base class for errors related to the ALH protocol
	"""
	pass

class ALHProtocolException(ALHException):
	def __init__(self, msg):
		if msg.endswith(self.TERMINATOR):
			msg = msg[:-len(self.TERMINATOR)].strip()
		super(Exception, self).__init__(msg)

class JunkInput(ALHProtocolException):
	TERMINATOR = "JUNK-INPUT\r\n"

class CorruptedData(ALHProtocolException): 
	TERMINATOR = "CORRUPTED-DATA\r\n"

class ALHRandomError(ALHException): pass

class CRCError(ALHException): pass

class TerminalError(IOError): pass

class ALHURLOpener(FancyURLopener):
	version = "vesna-alh-tools/1.0"

	def __init__(self):
		try:
			context = ssl._create_unverified_context()
		except AttributeError:
			context = None

		FancyURLopener.__init__(self, context=context)

	def prompt_user_passwd(self, host, realm):

		paths = [
				'alhrc',
				'/etc/alhrc',
			]

		home = os.environ.get('HOME')
		if home is not None:
			paths.append(os.path.join(home, '.alhrc'))

		for path in paths:
			try:
				f = open(path)
			except IOError:
				continue

			match = False
			user = None
			passwd = None

			for line in f:
				if line.startswith('#'):
					continue

				try:
					key, value = line.strip().split()
				except ValueError:
					continue

				if (key == 'Host'):
					match = (value == host)
					user = None
					passwd = None
				elif match and (key == 'User'):
					user = value
				elif match and (key == 'Password'):
					passwd = value

				if match and user and passwd:
					return (user, passwd)

		return FancyURLopener.prompt_user_passwd(self, host, realm)

class ALHProtocol:
	"""Base class for an ALH protocol service.

	This is an abstract class with some useful private methods.

	Implementations of this interface should override _get() and _post() methods.
	"""
	RETRIES = 5

	def get(self, resource, *args):
		"""Issue a GET request to the service.

		Raises an ALHException in case of an error.

		:param resource: resource to issue request to
		:param args: arbitrary string arguments for the request

		:return: the string reply from the resource handler
		"""
		return self._get(resource, *args)

	def post(self, resource, data, *args):
		"""Issue a POST request to the service

		Raises an ALHException in case of an error.

		:param resource: resource to issue request to
		:param data: POST data to attach to the request
		:param args: arbitrary string arguments for the request

		:return: the string reply from the resource handler
		"""
		return self._post(resource, data, *args)

	def _log_request(self, method, resource, args, data=None):
		log.info("%8s: %s?%s" % (method, resource, "".join(args)))
		if data is not None and len(data) > 4 and all(c in string.printable for c in data):
			log.info("    DATA: %s" % (data,))

	@staticmethod
	def _is_printable(resp):
		try:
			resp_ascii = resp.decode('ascii')
		except UnicodeDecodeError:
			return False

		return all(c in string.printable for c in resp_ascii)

	def _log_response(self, resp):
		if self._is_printable(resp):
			resp_ascii = resp.decode("ascii", "ignore").strip()
			log.info("response: %s" % (resp_ascii,))
		else:
			log.info("unprintable response (%d bytes)" % (len(resp),))

	def _send_with_retry(self, data):

		for retry in range(self.RETRIES):
			try:
				return self._send_with_error(data)
			except ALHException as e:
				if retry == self.RETRIES - 1:
					raise e
				else:
					log.exception("retrying (%d)" % (retry+1,))

	def _check_for_sneaky_error(self, resp):
		# This is extremely ugly. But since we don't have
		# currently any consistent way of specifying whether
		# a request failed or not, we check if the response
		# contains any strings that look like error messages.

		r = resp.decode('unicode_escape').lower()
		r = r.replace("bus errors  :", "")
		r = r.replace("   : 0 (error)", "")
		if "error" in r or "warning" in r:
			raise ALHRandomError(resp)


class ALHTerminal(ALHProtocol):
	"""ALH protocol implementation through a serial terminal.

	This implementation is used for testing and debugging when a sensor
	node is connected directly to a computer over a serial line.

	:param f: path to the character device of the terminal
	"""
	RESPONSE_TERMINATOR = "\r\nOK\r\n"

	def __init__(self, f):
		self.f = f

	def _send(self, data):
		self.f.write(data)

		resp = ""
		while not resp.endswith(self.RESPONSE_TERMINATOR):
			d = self.f.read()
			if d:
				resp += d
			else:
				raise TerminalError

		return resp[:-len(self.RESPONSE_TERMINATOR)]

	def _recover(self):
		self._send("\r\n" * 5)

	def _crc(self, data):
		return binascii.crc32(data)

	def _send_with_error(self, data):
		resp = self._send(data)
		if resp.endswith("JUNK-INPUT\r\n"):
			self._recover()
			raise JunkInput(resp)
		if resp.endswith("CORRUPTED-DATA\r\n"):
			raise CorruptedData(resp)

		self._check_for_sneaky_error(resp)
		self._log_response(resp)

		return resp

	def _get(self, resource, *args):
		self._log_request("GET", resource, args)

		arg = "".join(args)
		return self._send_with_retry("get %s?%s\r\n" % (resource, arg))

	def _post(self, resource, data, *args):
		self._log_request("POST", resource, args, data)

		arg = "".join(args)

		req = "post %s?%s\r\nlength=%d\r\n%s\r\n" % (
				resource, arg, len(data), data)

		crc = self._crc(req)

		req += "crc=%d\r\n" % crc

		return self._send_with_retry(req)

class ALHWeb(ALHProtocol):
	"""ALH protocol implementation through the HTTP infrastructure server.

	ALHWeb is typically used to access the coordinator of a ZigBee mesh network.

	If the API end-point is using basic authentication, you will be
	prompted for credentials on the command line.

	You can also save credentials into either a file named `.alhrc` in your
	home directory or `alhrc` in the current directory. Format of the file
	is as in the following example::

	    Host example.com
	    User <username>
	    Password <password>
	    # more Host, User, Password lines can follow

	:param base_url: base URL of the HTTP API (e.g. `https://crn.log-a-tec.eu/communicator`)
	:param cluster_id: numerical cluster id
	"""

	def __init__(self, base_url, cluster_id):
		self.base_url = base_url
		self.cluster_id = cluster_id
		self.opener = ALHURLOpener()

	def _send(self, url):
		f = self.opener.open(url)
		resp = f.read()

		# Raise an exception if we got anything else than a 200 OK
		if f.getcode() != 200:
			raise TerminalError(resp)

		return resp

	def _send_with_error(self, url):
		# loop until communication channel is free and our request
		# goes through.
		time_start = time.time()
		while True:
			resp = self._send(url)
			if resp != "ERROR: Communication in progress":
				break

			log.info("communicator is busy (have been waiting for %d s)" %
					(time.time() - time_start))

			time.sleep(1)

		self._check_for_sneaky_error(resp)
		
		self._log_response(resp)
		return resp

	def _get(self, resource, *args):
		self._log_request("GET", resource, args)

		arg = "".join(args)
		query = (
				('method', 'get'),
				('resource', '%s?%s' % (resource, arg)),
				('cluster', str(self.cluster_id)),
		)

		url = "%s?%s" % (self.base_url, urlencode(query))

		return self._send_with_retry(url)

	def _post(self, resource, data, *args):
		self._log_request("POST", resource, args, data)

		arg = "".join(args)
		query = (
				('method', 'post'),
				('resource', '%s?%s' % (resource, arg)),
				('content', '%s' % (data,)),
				('cluster', str(self.cluster_id)),
		)

		url = "%s?%s" % (self.base_url, urlencode(query))

		return self._send_with_retry(url)

class ALHProxy(ALHProtocol):
	"""ALH protocol implementation through an ALH proxy.

	This implementation forwards arbitrary ALH requests through the "nodes"
	resource on an ALH service used as a proxy.

	ALHProxy is typically used to access nodes on the ZigBee mesh network behind
	the coordinator.

	:param alhproxy: ALH implementation used as a proxy
	:param addr: ZigBee address of the node to forward requests to
	"""
	def __init__(self, alhproxy, addr):
		self.alhproxy = alhproxy
		self.addr = addr

	def _recover_remote(self):
		self.alhproxy.post("radio/noderesetparser", "1", "%d" % (self.addr,))

	def _check_for_junk_state(self, message):
		g = re.search("NODES:Node ([0-9]+) parser is in junk state\r\nERROR", message)
		if g:
			assert(int(g.group(1)) == self.addr)
			self._recover_remote()

	def _get(self, resource, *args):
		try:
			return self.alhproxy.get("nodes", "%d/%s?" % (self.addr, resource), *args)
		except ALHRandomError as e:
			self._check_for_junk_state(str(e))
			raise

	def _post(self, resource, data, *args):
		try:
			response = self.alhproxy.post("nodes", data, "%d/%s?" % (self.addr, resource), *args)
		except ALHRandomError as e:
			self._check_for_junk_state(str(e))
			raise

		# For POST requests, coordinator adds some string at the start
		# of the response.

		# Clean it up here, so that responses via proxy are identical
		# to responses with direct connection.
		response = re.sub("^Node #%d return;" % (self.addr,), "", response)
		return response



#!/usr/bin/python

import os
from setuptools import setup

def get_long_description():
	return open(os.path.join(os.path.dirname(__file__), "README.rst")).read()

setup(name='vesna-alhtools',
      version='1.0.3',
      description='Tools for talking the VESNA almost-like-HTTP protocol',
      license='GPL',
      long_description=get_long_description(),
      author='Tomaz Solc',
      author_email='tomaz.solc@ijs.si',
      url='https://github.com/avian2/vesna-alh-tools',

      packages = [ 'vesna', 'vesna.alh', 'vesna.cdf' ],

      namespace_packages = [ 'vesna' ],

      scripts = [ 'scripts/alh-reprogram',
	      'scripts/alh-map',
	      'scripts/alh-tx-test',
	      'scripts/alh-endpoint-server',
	      'scripts/alh-measure-rssi' ],

      install_requires = [ 'vesna-spectrumsensor', 'numpy', 'python-dateutil', 'lxml' ],

      test_suite = 'test',
)



import itertools
import re
import time

from vesna.alh import CRCError

class Device:
	"""A signal generation device.

	A particular hardware model can have one or more physical signal generation devices, each of which
	can support one or more configurations.
	"""
	def __init__(self, id, name):
		"""Create a new device.

		id -- numeric device id.
		name -- string with a human readable name of the device.
		"""
		self.id = id
		self.name = name

class DeviceConfig: 
	"""Configuration for a signal generation device.

	The set of possible configurations for a device is usually hardware-dependent (i.e. a
	configuration usually reflects physical hardware settings). A configuration defines the usable
	frequency and power range.
	"""
	def __init__(self, id, name, device):
		"""Create a new device configuration.

		id -- numeric configuration id.
		name -- string with a human readable name of the configuration.
		device -- Device object to which this configuration applies.
		"""
		self.id = id
		self.name = name
		self.device = device

	def ch_to_hz(self, ch):
		"""Convert channel number to center frequency in hertz."""
		assert ch >= 0
		assert ch < self.num

		return self.base + self.spacing * ch

	def get_start_hz(self):
		"""Return the lowest settable frequency."""
		return self.ch_to_hz(0)

	def get_stop_hz(self):
		"""Return the highest settable frequency."""
		return self.ch_to_hz(self.num - 1)

	def covers(self, f_hz, power_dbm):
		"""Return true if this configuration can support the given frequency and power.

		:param f_hz: transmission frequency in hertz
		:param power_dbm: transmission power in dBm
		"""
		return f_hz >= self.get_start_hz() and f_hz <= self.get_stop_hz() and \
				power_dbm >= self.min_power and power_dbm <= self.max_power

	def get_tx_config(self, f_hz, power_dbm):
		"""Return the transmission configuration for the given frequency and power.

		:param f_hz: transmission frequency in hertz
		:param power_dbm: transmission power in dBm
		"""
		assert self.covers(f_hz, power_dbm)

		f_ch = int(round((f_hz - self.base) / self.spacing))

		return TxConfig(self, f_ch, power_dbm)

	def __str__(self):
		return "channel config %d,%d: %10d - %10d Hz" % (
				self.device.id, self.id, self.get_start_hz(), self.get_stop_hz())

class TxConfig:
	"""Transmission configuration for a signal generation device.

	:param config: :py:class:`DeviceConfig` device configuration object to use
	:param f_ch: frequency channel for transmission
	:param power_db: power level for transmission
	"""

	def __init__(self, config, f_ch, power_dbm):
		assert f_ch >= 0
		assert f_ch < config.num
		assert power_dbm >= config.min_power
		assert power_dbm <= config.max_power

		self.config = config
		self.f_ch = f_ch
		self.power_dbm = power_dbm

class ConfigList:
	"""List of devices and device configurations supported by attached hardware."""

	def __init__(self):
		"""Create a new list."""
		self.configs = []
		self.devices = []

	def _add_device(self, device):
		self.devices.append(device)

	def _add_config(self, config):
		self.configs.append(config)

	def get_config(self, device_id, config_id):
		"""Return the specified device configuration.

		:param device_id: numeric device id, as returned by the `list` command
		:param config_id: numeric configuration id, as returned by the `list` command
		"""
		for config in self.configs:
			if config.id == config_id and config.device.id == device_id:
				return config

		return None

	def get_tx_config(self, f_hz, power_dbm, name=None):
		"""Return best transmission configuration for specified requirements.

		:param f_hz: transmission frequency
		:param power_dbm: transmission power
		:param name: optional required sub-string in device configuration name
		"""

		candidates = []

		for config in self.configs:
			if name and name not in config.name:
				continue

			if not config.covers(f_hz, power_dbm):
				continue

			candidates.append(config)

		# pick fastest matching config
		candidates.sort(key=lambda x:x.time, reverse=True)

		if candidates:
			return candidates[0].get_tx_config(f_hz, power_dbm)
		else:
			return None

	def __str__(self):
		lines = []
		for device in self.devices:
			lines.append("device %d: %s" % (device.id, device.name))
			for config in self.configs:
				if config.device is device:
					lines.append("  channel config %d,%d: %s" % (device.id, config.id, config.name))
					lines.append("    base: %d Hz" % (config.base,))
					lines.append("    spacing: %d Hz" % (config.spacing,))
					lines.append("    bw: %d Hz" % (config.bw,))
					lines.append("    num: %d" % (config.num,))
					lines.append("    power: %d ... %d dBm" % (config.min_power, config.max_power))
					lines.append("    time: %d ms" % (config.time,))

		return '\n'.join(lines)

class SignalGeneratorProgram:
	"""Describes a single signal generation task.

	:param tx_config: transmission configuration to use
	:param time_start: time to start the task (UNIX timestamp)
	:param time_duration: duration of the task in seconds
	"""

	def __init__(self, tx_config, time_start, time_duration):
		self.tx_config = tx_config
		self.time_start = time_start
		self.time_duration = time_duration

class SignalGenerator:
	"""ALH node acting as a signal generator.

	:param alh: ALH implementation used to communicate with the node
	"""
	MAX_TIME_ERROR = 2.0

	def __init__(self, alh):
		self.alh = alh

	def program(self, program):
		"""Send the given signal generation program to the node.

		:param program: a :py:class:`SignalGeneratorProgram` object
		"""
		return self.program_list([program])

	def program_list(self, program_list):
		"""Send several signal generator programs to the node.

		:param program_list: a list of :py:class:`SignalGeneratorProgram` objects
		"""
		time_before = time.time()

		data_list = []

		for program in program_list:

			relative_time = int(program.time_start - time_before)
			if relative_time < 0:
				raise Exception("Start time can't be in the past")

			data_list.append("in %d sec for %d sec with dev %d conf %d channel %d power %d" % (
					relative_time,
					program.time_duration,
					program.tx_config.config.device.id,
					program.tx_config.config.id,
					program.tx_config.f_ch,
					program.tx_config.power_dbm))

		self.alh.post("generator/program", "\n".join(data_list))

		time_after = time.time()

		time_error = time_after - time_before
		if time_error > self.MAX_TIME_ERROR:
			raise Exception("Programming time error %.1f s > %.1fs" % 
					(time_error, self.MAX_TIME_ERROR))

	def get_config_list(self):
		"""Query and return the list of supported device configurations.

		:return: a :py:class:`ConfigList` object
		"""
		config_list = ConfigList()

		device = None
		config = None

		description = self.alh.get("generator/deviceConfigList")
		description_ascii = description.decode('ascii')
		configs_left = 0
		state = 0
		for line in description_ascii.split("\n"):
			g = re.match("dev #([0-9]+), (.+), ([0-9]+) configs:", line)
			if state == 0 and g:
				device = Device(int(g.group(1)), g.group(2))
				config_list._add_device(device)
				configs_left = int(g.group(3))
				state = 1
				continue

			g = re.match("  cfg #([0-9]+): (.+):", line)
			if state == 1 and g:
				config = DeviceConfig(int(g.group(1)), g.group(2), device)
				state = 2

				continue

			g = re.match("     base: ([0-9]+) Hz, spacing: ([0-9]+) Hz, bw: ([0-9]+) Hz, channels: ([0-9]+), min power: ([0-9-]+) dBm, max power: ([0-9-]+) dBm, time: ([0-9]+) ms", line)
			if state == 2 and g:
				config.base = int(g.group(1))
				config.spacing = int(g.group(2))
				config.bw = int(g.group(3))
				config.num = int(g.group(4))
				config.min_power = int(g.group(5))
				config.max_power = int(g.group(6))
				config.time = int(g.group(7))

				config_list._add_config(config)

				configs_left -= 1
				if configs_left < 0:
					raise CRCError
				elif configs_left == 0:
					state = 0
				else:
					state = 1

				continue

		if configs_left != 0:
			raise CRCError

		return config_list



import datetime
import json
import logging
import os.path
import sys
import time
import uuid
import vesna.alh
import vesna.alh.spectrumsensor
import vesna.alh.signalgenerator
import vesna.alh.common

log = logging.getLogger(__name__)

def isstring(v):
	if sys.version_info[0] >= 3:
		return isinstance(v, str)
	else:
		return isinstance(v, str) or isinstance(v, unicode)

def force_list(v):
	if v is None:
		return []
	elif isstring(v):
		return [ v ]
	else:
		return v

class CDFError(Exception): pass

class CDFInterfererProgram:
	def __init__(self, center_hz, power_dbm, start_time, end_time, device_id=None, config_id=None):
		self.center_hz = center_hz
		self.power_dbm = power_dbm
		self.device_id = device_id
		self.config_id = config_id

		assert end_time > start_time

		self.start_time = start_time
		self.end_time = end_time

class CDFInterferer:
	def __init__(self, device):
		self.device = device
		self.programs = []

	def add_program(self, program):
		self.programs.append(program)

class CDFDevice:
	def __init__(self, base_url, cluster_id, addr):
		self.base_url = base_url
		self.cluster_id = cluster_id
		self.addr = addr

	def key(self):
		return (self.base_url, self.cluster_id, self.addr)

	def __str__(self):
		return "<CDFDevice base_url=%s cluster_id=%d addr=%d>" % (
				self.base_url, self.cluster_id, self.addr)

class CDFAuthor:
	def __init__(self, name, email, address=None, phone=None, institution=None):
		self.name = name
		self.email = email
		self.address = force_list(address)
		self.phone = force_list(phone)
		self.institution = force_list(institution)

class CDFDocument:
	def __init__(self, description=None, bibtex=None):
		self.description = force_list(description)
		self.bibtex = force_list(bibtex)

class CDFExperimentIteration:
	def __init__(self, slot_id=10):
		self.slot_id = slot_id

		self.sensors = []
		self.interferers = []

		self.start_time = None
		self.end_time = None

		self.tracefiles = []

class CDFExperimentSensor:
	def __init__(self, sensor):
		self.sensor = sensor

class CDFExperimentInterferer:
	def __init__(self, generator):
		self.generator = generator

		self.program_list = []

class CDFExperiment:
	def __init__(self, title, summary, related_experiments, notes, methodology=None,
			tag=None, release_date=None, authors=None, documentation=None, devices=None, 
			interferers=None):

		self.title = title

		if tag is not None:
			self.tag = tag
		else:
			self.tag = "vesna-alh-tools-" + str(uuid.uuid4())

		self.authors = force_list(authors)

		self.release_date = release_date

		self.summary = summary

		self.methodology = force_list(methodology)

		self.documentation = force_list(documentation)

		self.related_experiments = related_experiments

		self.notes = force_list(notes)

		self.start_hz = None
		self.stop_hz = None
		self.step_hz = None

		self.duration = None

		self.devices = force_list(devices)
		self.interferers = force_list(interferers)

		self.iterations = []

	def set_frequency_range(self, start_hz, stop_hz, step_hz):
		self.start_hz = start_hz
		self.stop_hz = stop_hz
		self.step_hz = step_hz

	def set_duration(self, duration):
		self.duration = duration

	def add_author(self, author):
		self.authors.append(author)

	def add_document(self, document):
		self.documentation.append(document)

	def add_device(self, device):
		self.devices.append(device)

	def add_interferer(self, interferer):
		self.interferers.append(interferer)

	def iter_all_devices(self):
		for device in self.devices:
			yield device

		for interferer in self.interferers:
			yield interferer.device

	def _get_coordinators(self):
		coordinators = {}

		for device in self.iter_all_devices():
			args = (device.base_url, device.cluster_id)
			if args not in coordinators:
				coordinator = vesna.alh.ALHWeb(*args)

				coordinator.post("prog/firstCall", "1")

				coordinators[args] = coordinator

		return coordinators

	def _get_nodes(self):
		coordinators = self._get_coordinators()

		nodes = {}

		for device in self.iter_all_devices():
			if device.key() not in nodes:
				coordinator = coordinators[device.base_url, device.cluster_id]

				node = vesna.alh.ALHProxy(coordinator, device.addr)

				node.post("prog/firstCall", "1")

				nodes[device.key()] = node
			else:
				raise CDFError("Device %s used more than once" % device)

		return nodes

	def run(self, iteration):
		sensors = iteration.sensors
		interferers = iteration.interferers

		nodes = self._get_nodes()

		start_time = time.time() + 5.0 * len(nodes)
		end_time = start_time + self.duration

		iteration.start_time = datetime.datetime.fromtimestamp(start_time)
		iteration.end_time = datetime.datetime.fromtimestamp(end_time)

		for device in self.devices:
			node = nodes[device.key()]

			sensor = CDFExperimentSensor(vesna.alh.spectrumsensor.SpectrumSensor(node))

			config_list = sensor.sensor.get_config_list()

			sweep_config = config_list.get_sweep_config(
					start_hz=self.start_hz,
					stop_hz=self.stop_hz,
					step_hz=self.step_hz)

			if sweep_config is None:
				raise CDFError("Device %s cannot scan desired frequency range" % device)

			sensor.program = vesna.alh.spectrumsensor.SpectrumSensorProgram(
					sweep_config, 
					start_time,
					end_time - start_time,
					slot_id=iteration.slot_id)

			sensors.append(sensor)

		for interferer in self.interferers:
			node = nodes[interferer.device.key()]

			einterferer = CDFExperimentInterferer(
					vesna.alh.signalgenerator.SignalGenerator(node))

			config_list = einterferer.generator.get_config_list()

			for program in interferer.programs:
				tx_config = config_list.get_tx_config(
						f_hz=program.center_hz,
						power_dbm=program.power_dbm)

				if tx_config is None:
					raise CDFError("Device %s cannot transmit at desired "
							"frequency range" % interferer.device)

				einterferer.program_list.append(
						vesna.alh.signalgenerator.SignalGeneratorProgram(
								tx_config,
								start_time + program.start_time,
								program.end_time - program.start_time))

			interferers.append(einterferer)

		for sensor in sensors:
			sensor.sensor.program(sensor.program)
		for interferer in interferers:
			interferer.generator.program_list(interferer.program_list)

		for sensor in sensors:
			while not sensor.sensor.is_complete(sensor.program):
				log.info("waiting")
				time.sleep(2)

				if time.time() > (end_time + 30):
					raise Exception("Something went wrong")

			log.info("experiment is finished. retrieving data.")

			sensor.result = sensor.sensor.retrieve(sensor.program)

		self.iterations.append(iteration)
